searchData={"items":[{"type":"module","title":"Integrator","doc":"A library for solving non-stiff ordinary differential equations (ODEs).\n\n`Integrator` uses either the Dormand-Prince 4/5 Runge Kutta algorithm, or the Bogacki-Shampine 2/3\nRunge Kutta algorithm.  It is intended that the user only needs to call `Integrator.integrate/4`, and\nthe `Integrator.AdaptiveStepsize` and `Integrator.RungeKutta` modules are only exposed for advanced\nusers who want to use the underlying algorithms directly.","ref":"Integrator.html"},{"type":"function","title":"Integrator.integrate/4","doc":"Integrates an ODE function using either the Dormand-Prince45 method or the Bogacki-Shampine23 method.","ref":"Integrator.html#integrate/4"},{"type":"function","title":"Options - Integrator.integrate/4","doc":"* `:initial_step` - The initial stepsize. If not provided, a stepsize will be chosen automatically. Can be a float\n  or a Nx tensor.\n\n* `:integrator` - The integrator to use. Currently only :ode45 and :ode23 are supported, which correspond to\n  `Integrator.RungeKutta.DormandPrince45` and `Integrator.RungeKutta.BogackiShampine23`, respectively. The default value is `:ode45`.\n\n\n\n#","ref":"Integrator.html#integrate/4-options"},{"type":"function","title":"Additional Options - Integrator.integrate/4","doc":"Also see the options for these functions which are passed through:\n\n* `Integrator.NonLinearEqnRoot.find_zero/4`\n* `Integrator.AdaptiveStepsize.integrate/10`","ref":"Integrator.html#integrate/4-additional-options"},{"type":"function","title":"Integrator.options_schema_integrator_only/0","doc":"","ref":"Integrator.html#options_schema_integrator_only/0"},{"type":"module","title":"Integrator.AdaptiveStepsize","doc":"Integrates a set of ODEs with an adaptive timestep.","ref":"Integrator.AdaptiveStepsize.html"},{"type":"function","title":"Integrator.AdaptiveStepsize.abs_rel_norm_opts/1","doc":"Gets the default values used by the absolute-relative norm; e.g., `abs_tol`, `rel_tol`, and\n`norm_control`","ref":"Integrator.AdaptiveStepsize.html#abs_rel_norm_opts/1"},{"type":"function","title":"Integrator.AdaptiveStepsize.elapsed_time_ms/1","doc":"Returns the total elapsed time for the integration (in milleseconds)","ref":"Integrator.AdaptiveStepsize.html#elapsed_time_ms/1"},{"type":"function","title":"Integrator.AdaptiveStepsize.integrate/10","doc":"Integrates a set of ODEs.","ref":"Integrator.AdaptiveStepsize.html#integrate/10"},{"type":"function","title":"Options - Integrator.AdaptiveStepsize.integrate/10","doc":"* `:abs_tol` (`t:term/0`) - The absolute tolerance used when computing the absolute relative norm. Defaults to 1.0e-06 in the Nx type that's been specified.\n\n* `:event_fn` - A 2 arity function which determines whether an event has occured.  If so, the integration is halted. The default value is `nil`.\n\n* `:max_number_of_errors` (`t:integer/0`) - The maximum number of permissible errors before the integration is halted. The default value is `5000`.\n\n* `:max_step` (`t:term/0`) - The default max time step.  The default value is determined by the start and end times.\n\n* `:norm_control` (`t:boolean/0`) - Indicates whether norm control is to be used when computing the absolute relative norm. The default value is `true`.\n\n* `:output_fn` - A 2 arity function which is called at each output point. The default value is `nil`.\n\n* `:refine` - Indicates the number of additional interpolated points. `1` means no interpolation; `2` means one\n  additional interpolated point; etc. `:fixed_times` means that the output times are fixed. The default value is `4`.\n\n* `:rel_tol` (`t:term/0`) - The relative tolerance used when computing the absolute relative norm. Defaults to 1.0e-03 in the Nx type that's been specified.\n\n* `:speed` - `:no_delay` means to simulate as fast as possible. `1.0` means real time, `2.0` means twice as fast as real time,\n  `0.5` means half as fast as real time, etc. The default value is `:no_delay`.\n\n* `:store_results?` (`t:boolean/0`) - Indicates whether or not to store the results of the integration. The default value is `true`.\n\n\n\n#","ref":"Integrator.AdaptiveStepsize.html#integrate/10-options"},{"type":"function","title":"Additional Options - Integrator.AdaptiveStepsize.integrate/10","doc":"Also see the options for the `Integrator.NonLinearEqnRoot.find_zero/4` which are passed\ninto `integrate/10`.\n\nOriginally adapted from the Octave\n[integrate_adaptive.m](https://github.com/gnu-octave/octave/blob/default/scripts/ode/private/integrate_adaptive.m)\n\nSee [Wikipedia](https://en.wikipedia.org/wiki/Adaptive_stepsize)","ref":"Integrator.AdaptiveStepsize.html#integrate/10-additional-options"},{"type":"function","title":"Integrator.AdaptiveStepsize.option_keys/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#option_keys/0"},{"type":"function","title":"Integrator.AdaptiveStepsize.options_schema/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#options_schema/0"},{"type":"function","title":"Integrator.AdaptiveStepsize.options_schema_adaptive_stepsize_only/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#options_schema_adaptive_stepsize_only/0"},{"type":"function","title":"Integrator.AdaptiveStepsize.starting_stepsize/7","doc":"Computes a good initial timestep for an ODE solver of order `order`\nusing the algorithm described in the reference below.\n\nThe input argument `ode_fn`, is the function describing the differential\nequations, `t0` is the initial time, and `x0` is the initial\ncondition.  `abs_tol` and `rel_tol` are the absolute and relative\ntolerance on the ODE integration.\n\nOriginally based on the Octave\n[`starting_stepsize.m`](https://github.com/gnu-octave/octave/blob/default/scripts/ode/private/starting_stepsize.m).\n\nReference:\n\nE. Hairer, S.P. Norsett and G. Wanner,\n\"Solving Ordinary Differential Equations I: Nonstiff Problems\",\nSpringer.","ref":"Integrator.AdaptiveStepsize.html#starting_stepsize/7"},{"type":"type","title":"Integrator.AdaptiveStepsize.event_fn_t/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:event_fn_t/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.integration_status/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:integration_status/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.options_t/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:options_t/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.output_fn_t/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:output_fn_t/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.refine_strategy/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:refine_strategy/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.speed/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:speed/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.t/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:t/0"},{"type":"module","title":"Integrator.AdaptiveStepsize.ComputedStep","doc":"The results of the computation of an individual Runge-Kutta step","ref":"Integrator.AdaptiveStepsize.ComputedStep.html"},{"type":"type","title":"Integrator.AdaptiveStepsize.ComputedStep.t/0","doc":"","ref":"Integrator.AdaptiveStepsize.ComputedStep.html#t:t/0"},{"type":"module","title":"Integrator.MultiIntegrator","doc":"Integrates multiple simulations that are tied together somehow, such as a bouncing ball\nin the `ballode.m` example.","ref":"Integrator.MultiIntegrator.html"},{"type":"function","title":"Integrator.MultiIntegrator.all_output_data/2","doc":"Collates the simulation output from all of the integrations","ref":"Integrator.MultiIntegrator.html#all_output_data/2"},{"type":"function","title":"Integrator.MultiIntegrator.integrate/7","doc":"Integrates multiple times, with a transition function handling the junction between integrations","ref":"Integrator.MultiIntegrator.html#integrate/7"},{"type":"function","title":"Options - Integrator.MultiIntegrator.integrate/7","doc":"See the options for these functions which are passed through:\n\n* `Integrator.NonLinearEqnRoot.find_zero/4`\n* `Integrator.AdaptiveStepsize.integrate/10`\n* `Integrator.integrate/4`","ref":"Integrator.MultiIntegrator.html#integrate/7-options"},{"type":"type","title":"Integrator.MultiIntegrator.integration_status/0","doc":"","ref":"Integrator.MultiIntegrator.html#t:integration_status/0"},{"type":"type","title":"Integrator.MultiIntegrator.t/0","doc":"","ref":"Integrator.MultiIntegrator.html#t:t/0"},{"type":"type","title":"Integrator.MultiIntegrator.transition_fn_t/0","doc":"","ref":"Integrator.MultiIntegrator.html#t:transition_fn_t/0"},{"type":"module","title":"Integrator.NonLinearEqnRoot","doc":"Finds the roots (i.e., zeros) of a non-linear equation.\nBased on [`fzero.m`](https://github.com/gnu-octave/octave/blob/default/scripts/optimization/fzero.m)\nfrom Octave.\n\nThis is essentially the ACM algorithm 748: Enclosing Zeros of Continuous Functions\ndue to Alefeld, Potra and Shi, ACM Transactions on Mathematical Software, Vol. 21,\nNo. 3, September 1995. Although the workflow is the same, the structure of\nthe algorithm has been transformed non-trivially; also, the algorithm has also been\nslightly modified.","ref":"Integrator.NonLinearEqnRoot.html"},{"type":"function","title":"Integrator.NonLinearEqnRoot.bracket_fx/1","doc":"","ref":"Integrator.NonLinearEqnRoot.html#bracket_fx/1"},{"type":"function","title":"Integrator.NonLinearEqnRoot.bracket_x/1","doc":"","ref":"Integrator.NonLinearEqnRoot.html#bracket_x/1"},{"type":"function","title":"Integrator.NonLinearEqnRoot.find_zero/4","doc":"Finds a zero for a function in an interval `[a, b]` (if the 2nd argument is a list) or\nin the vicinity of `a` (if the 2nd argument is a float).","ref":"Integrator.NonLinearEqnRoot.html#find_zero/4"},{"type":"function","title":"Options - Integrator.NonLinearEqnRoot.find_zero/4","doc":"* `:max_iterations` (`t:integer/0`) - The maximum allowed number of iterations when finding a root. The default value is `1000`.\n\n* `:max_fn_eval_count` (`t:integer/0`) - The maximum allowed number of function evaluations when finding a root. The default value is `1000`.\n\n* `:type` - The Nx type. The default value is `:f64`.\n\n* `:machine_eps` (`t:float/0`) - The machine epsilon. Defaults to Nx.constants.epsilon/1 for this Nx type.\n\n* `:tolerance` (`t:float/0`) - The tolerance for the convergence when finding a root. Defaults to Nx.Constants.epsilon/1 for this Nx type.\n\n* `:nonlinear_eqn_root_output_fn` (`t:term/0`) - An output function to call so intermediate results can be retrieved when finding a root. The default value is `nil`.","ref":"Integrator.NonLinearEqnRoot.html#find_zero/4-options"},{"type":"function","title":"Integrator.NonLinearEqnRoot.option_keys/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#option_keys/0"},{"type":"function","title":"Integrator.NonLinearEqnRoot.options_schema/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#options_schema/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.convergence_status/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:convergence_status/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.interpolation_type/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:interpolation_type/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.iter_type/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:iter_type/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.options_t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:options_t/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.output_fn_t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:output_fn_t/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.search_for_2nd_point_t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:search_for_2nd_point_t/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:t/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.zero_fn_t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:zero_fn_t/0"},{"type":"behaviour","title":"Integrator.RungeKutta","doc":"A behaviour that Runge-Kutta algorithms must implement. Currently, `Integrator.DormandPrince45`\nand `Integrator.BogackiShampine23` implement this behaviour.\n\nSee the [list of Runge-Kutta methods](https://en.wikipedia.org/wiki/List_of_Runge%E2%80%93Kutta_methods)","ref":"Integrator.RungeKutta.html"},{"type":"callback","title":"Integrator.RungeKutta.integrate/6","doc":"Integrates an ODE function","ref":"Integrator.RungeKutta.html#c:integrate/6"},{"type":"callback","title":"Integrator.RungeKutta.interpolate/4","doc":"Interpolates using the method that is suitable for this particular Runge-Kutta method","ref":"Integrator.RungeKutta.html#c:interpolate/4"},{"type":"callback","title":"Integrator.RungeKutta.order/0","doc":"The order of this Runge-Kutta method","ref":"Integrator.RungeKutta.html#c:order/0"},{"type":"type","title":"Integrator.RungeKutta.interpolate_fn_t/0","doc":"","ref":"Integrator.RungeKutta.html#t:interpolate_fn_t/0"},{"type":"type","title":"Integrator.RungeKutta.ode_fn_t/0","doc":"","ref":"Integrator.RungeKutta.html#t:ode_fn_t/0"},{"type":"type","title":"Integrator.RungeKutta.stepper_fn_t/0","doc":"","ref":"Integrator.RungeKutta.html#t:stepper_fn_t/0"},{"type":"module","title":"Integrator.RungeKutta.BogackiShampine23","doc":"Bogacki-Shampine method of third order.  For the definition of this method see\n[Wikipedia](http://en.wikipedia.org/wiki/List_of_Runge%E2%80%93Kutta_methods)\n\nOriginally based on Octave [`runge_kutta_23.m`](https://github.com/gnu-octave/octave/blob/default/scripts/ode/private/runge_kutta_23.m)","ref":"Integrator.RungeKutta.BogackiShampine23.html"},{"type":"function","title":"Integrator.RungeKutta.BogackiShampine23.integrate/6","doc":"Solves a set of non-stiff Ordinary Differential Equations (non-stiff ODEs) with the well-known\nexplicit Bogacki-Shampine method of order 3.\n\nSee [Wikipedia here](https://en.wikipedia.org/wiki/List_of_Runge%E2%80%93Kutta_methods#Bogacki%E2%80%93Shampine)\nand [here](https://en.wikipedia.org/wiki/Bogacki%E2%80%93Shampine_method)","ref":"Integrator.RungeKutta.BogackiShampine23.html#integrate/6"},{"type":"function","title":"Integrator.RungeKutta.BogackiShampine23.interpolate/4","doc":"Performs a Hermite cubic interpolation when using BogackiShampine23 via\n`Utils.hermite_cubic_interpolation/4`","ref":"Integrator.RungeKutta.BogackiShampine23.html#interpolate/4"},{"type":"function","title":"Integrator.RungeKutta.BogackiShampine23.order/0","doc":"Returns the order of this Runge-Kutta method (which is 3)","ref":"Integrator.RungeKutta.BogackiShampine23.html#order/0"},{"type":"module","title":"Integrator.RungeKutta.DormandPrince45","doc":"Integrates and interpolates a system of ODEs with a given initial condition `x` from `t`\nto `t + dt` with the Dormand-Prince method. For the definition of this method see\n[Wikipedia](http://en.wikipedia.org/wiki/Dormand%E2%80%93Prince_method).\n\nOriginally based on [`runge_kutta_45_dorpri.m`](https://github.com/gnu-octave/octave/blob/default/scripts/ode/private/runge_kutta_45_dorpri.m)\nfrom Octave.\n\nIt uses six function evaluations to calculate fourth and fifth-order accurate solutions.\nThe difference between these solutions is then taken to be the error of the (fourth-order) solution.\nThis error estimate is very convenient for adaptive stepsize integration algorithms.\nThe Dormand–Prince method has seven stages, but it uses only six function evaluations per step\nbecause it has the FSAL (First Same As Last) property: the last stage is evaluated at the same\npoint as the first stage of the next step. Dormand and Prince chose the coefficients of their\nmethod to minimize the error of the fifth-order solution. This is the main difference with the\n[Fehlberg method](https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta%E2%80%93Fehlberg_method), which\nwas constructed so that the fourth-order solution has a small error.\n\nSee the [Octave docs for ode45](https://docs.octave.org/interpreter/Matlab_002dcompatible-solvers.html#XREFode45)","ref":"Integrator.RungeKutta.DormandPrince45.html"},{"type":"function","title":"Integrator.RungeKutta.DormandPrince45.integrate/6","doc":"Integrates a system of ODEs with\n[Dormand-Prince]](http://en.wikipedia.org/wiki/Dormand%E2%80%93Prince_method).\n\nReference: Hairer, Ernst; Nørsett, Syvert Paul; Wanner, Gerhard (2008),\nSolving ordinary differential equations I: Nonstiff problems,\nBerlin, New York: Springer-Verlag, ISBN 978-3-540-56670-0","ref":"Integrator.RungeKutta.DormandPrince45.html#integrate/6"},{"type":"function","title":"Integrator.RungeKutta.DormandPrince45.interpolate/4","doc":"Performs a 4th order Hermite interpolation when interpolating with DormandPrince45\nusing `Utils.hermite_quartic_interpolation/4`","ref":"Integrator.RungeKutta.DormandPrince45.html#interpolate/4"},{"type":"function","title":"Integrator.RungeKutta.DormandPrince45.order/0","doc":"Returns the order of this Runge-Kutta method (which is 5)","ref":"Integrator.RungeKutta.DormandPrince45.html#order/0"},{"type":"module","title":"Integrator.SampleEqns","doc":"Functions to be used in testing","ref":"Integrator.SampleEqns.html"},{"type":"function","title":"Integrator.SampleEqns.euler_equations/2","doc":"The Euler equations of a rigid body without external forces. This is a standard test\nproblem proposed by Krogh for solvers intended for nonstiff problems [see below].\nBased on \"rigidode.m\" from Matlab/Octave.  The analytical solutions are Jacobian\nelliptic functions.\n\nSee [`rigidode.m`](http://www.ece.northwestern.edu/local-apps/matlabhelp/techdoc/math_anal/diffeq8.html)\nin Matlab.\n\nShampine, L. F., and M. K. Gordon, Computer Solution of Ordinary Differential Equations,\nW.H. Freeman & Co., 1975","ref":"Integrator.SampleEqns.html#euler_equations/2"},{"type":"function","title":"Integrator.SampleEqns.falling_particle/2","doc":"Simulates a point mass or particle falling through pass affected by gravity.  Used for comparisons\nwith the Matlab/Octave `ballode.m` routine","ref":"Integrator.SampleEqns.html#falling_particle/2"},{"type":"function","title":"Integrator.SampleEqns.van_der_pol_fn/2","doc":"From [octave](https://octave.sourceforge.io/octave/function/ode45.html) b\n\n\nfvdp = @(t,x) [x(2); (1 - x(1)^2) * x(2) - x(1)];","ref":"Integrator.SampleEqns.html#van_der_pol_fn/2"},{"type":"module","title":"Integrator.Utils","doc":"Various utility functions used in `Integrator`","ref":"Integrator.Utils.html"},{"type":"function","title":"Integrator.Utils.columns_as_list/3","doc":"Returns the columns of a tensor as a list of vector tensors\n\nE.g., a tensor of the form:\n\n    ~M[\n      1  2  3  4\n      5  6  7  8\n      9 10 11 12\n    ]s8\n\nwill return\n\n  [\n    ~V[ 1  5   9 ]s8,\n    ~V[ 2  6  10 ]s8,\n    ~V[ 3  7  11 ]s8,\n    ~V[ 4  8  12 ]s8\n  ]","ref":"Integrator.Utils.html#columns_as_list/3"},{"type":"function","title":"Integrator.Utils.epsilon/1","doc":"Gets the epsilon for a particular Nx type","ref":"Integrator.Utils.html#epsilon/1"},{"type":"function","title":"Integrator.Utils.epsilon_nx/1","doc":"","ref":"Integrator.Utils.html#epsilon_nx/1"},{"type":"function","title":"Integrator.Utils.hermite_cubic_interpolation/4","doc":"Performs a 3rd order Hermite interpolation. Adapted from function `hermite_cubic_interpolation` in\n[runge_kutta_interpolate.m](https://github.com/gnu-octave/octave/blob/default/scripts/ode/private/runge_kutta_interpolate.m)\n\n\nSee [Wikipedia](https://en.wikipedia.org/wiki/Cubic_Hermite_spline)","ref":"Integrator.Utils.html#hermite_cubic_interpolation/4"},{"type":"function","title":"Integrator.Utils.hermite_quartic_interpolation/4","doc":"Performs a 4th order Hermite interpolation. Used by an ODE solver to interpolate the\nsolution at the time `t_out`. As proposed by Shampine in Lawrence, Shampine,\n\"Some Practical Runge-Kutta Formulas\", 1986.\n\nSee [hermite_quartic_interpolation function in Octave](https://github.com/gnu-octave/octave/blob/default/scripts/ode/private/runge_kutta_interpolate.m#L91).","ref":"Integrator.Utils.html#hermite_quartic_interpolation/4"},{"type":"function","title":"Integrator.Utils.kahan_sum/3","doc":"Implements the Kahan summation algorithm, also known as compensated summation.\nBased on this [code in Octave](https://github.com/gnu-octave/octave/blob/default/scripts/ode/private/kahan.m).\nThis is really a private function, but is made public so the docs are visible.\n\nThe algorithm significantly reduces the numerical error in the total\nobtained by adding a sequence of finite precision floating point numbers\ncompared to the straightforward approach.  For more details\nsee [this Wikipedia entry](http://en.wikipedia.org/wiki/Kahan_summation_algorithm).\nThis function is called by AdaptiveStepsize.integrate to better catch\nequality comparisons.\n\nThe first input argument is the variable that will contain the summation.\nThis variable is also returned as the first output argument in order to\nreuse it in subsequent calls to `kahan_sum/3` function.\n\nThe second input argument contains the compensation term and is returned\nas the second output argument so that it can be reused in future calls of\nthe same summation.\n\nThe third input argument `term` is the variable to be added to `sum`.","ref":"Integrator.Utils.html#kahan_sum/3"},{"type":"function","title":"Integrator.Utils.sign/1","doc":"Returns the sign of the tensor as -1 or 1 (or 0 for zero tensors)","ref":"Integrator.Utils.html#sign/1"},{"type":"function","title":"Integrator.Utils.unique/1","doc":"Given a list of elements, create a new list with only the unique elements from the list","ref":"Integrator.Utils.html#unique/1"},{"type":"function","title":"Integrator.Utils.vector_as_list/1","doc":"Converts a Nx vector into a list of 1-D tensors\n\nIs there an existing Nx way to do this?  If so, swap the usage of this function\nand then delete this\n\nNote that\n\n    vector |> Nx.as_list() |> Enum.map(& Nx.tensor(&1, type: Nx.type(vector)))\n\nseems to introduce potential precision issues","ref":"Integrator.Utils.html#vector_as_list/1"},{"type":"extras","title":"Integrator","doc":"# Integrator\n\nA numerical integrator written in Elixir for the solution of sets of non-stiff ordinary differential\nequations (ODEs).","ref":"readme.html"},{"type":"extras","title":"Installation - Integrator","doc":"The package can be installed by adding `integrator` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:integrator, \"~> 0.1\"},\n  ]\nend\n```\n\nThe docs can be found at  .","ref":"readme.html#installation"},{"type":"extras","title":"Description - Integrator","doc":"Two integrator options are available; `ode45` which is an adaptation of the [Octave\node45](https://octave.sourceforge.io/octave/function/ode45.html) and [Matlab\node45](https://www.mathworks.com/help/matlab/ref/ode45.html). The `ode45` integrator utilizes the\n[Dormand-Prince](https://en.wikipedia.org/wiki/Dormand%E2%80%93Prince_method) 4th/5th order Runge\nKutta algorithm.\n\n`ode23` is an adaptation of the [Octave\node23](https://octave.sourceforge.io/octave/function/ode23.html) and [Matlab\node23](https://www.mathworks.com/help/matlab/ref/ode23.html) The `ode23` integrator uses the\n[Bogacki-Shampine](https://en.wikipedia.org/wiki/Bogacki%E2%80%93Shampine_method) 3rd order Runge\nKutta algorithm.\n\nBoth `ode45` (which is the default integrator option) and `ode23` utilize an adaptive stepsize\nalgorithm for computing the integration time step.  The time step is computed based on the\nsatisfaction of a required error tolerance.\n\nThis library heavily leverages [Elixir Nx](https://github.com/elixir-nx/nx); many thanks to the\n[creators of `Nx`](https://github.com/elixir-nx/nx/graphs/contributors), as without it this library\nwould not have been possible. The [GNU Octave code](https://github.com/gnu-octave/octave) was also\nused heavily for inspiration and was used to generate numerical test cases for the Elixir versions\nof the algorithms.  Many thanks to [John W. Eaton](https://jweaton.org/) for his tremendous work on\nOctave. `Integrator` has been tested extensively during its development, and has a large and growing\ntest suite.","ref":"readme.html#description"},{"type":"extras","title":"Usage - Integrator","doc":"See the Livebook guides for detailed examples of usage. As a simple example, you can integrate the\nVan der Pol equation as defined in `Integrator.SampleEqns.van_der_pol_fn/2` from time 0 to 20 with an\nintial x value of `[0, 1]` via:\n\n```elixir\nt_initial = 0.0\nt_final = 20.0\nx_initial = Nx.tensor([0.0, 1.0])\nsolution = Integrator.integrate(&SampleEqns.van_der_pol_fn/2, [t_initial, t_final], x_initial)\n```\n\nThen, `solution.output_t` contains a list of output times, and `solution.output_x` contains a list\nof values of `x` at these corresponding times.\n\n![images/van_der_pol](images/van_der_pol.png)\n\nOptions exist for:\n- outputting simulation results dynamically via an output function (for applications\nsuch as plotting dynamically, or for animating while the simulation is underway)\n- generating simulation output at fixed times (such as at `t = 0.1, 0.2, 0.3`, etc.)\n- interpolating intermediate points via quartic Hermite interpolation (for `ode45`) or via cubic\nHermite interpolation (for `ode23`) \n- detecting termination events (such as collisions); see the Livebooks for details.\n- increasing the simulation fidelity (at the expense of simulation time) via absolute tolerance and\n  relative tolerance settings","ref":"readme.html#usage"},{"type":"extras","title":"So why should I care??? A tool to solve ODEs? WTF??? - Integrator","doc":"The basic gist of the project is that it is a tool in Elixir (that leverages [Nx](https://github.com/elixir-nx)) \nto numerically solve sets of ordinary differential equations (ODEs).  Science and engineering \nproblems typically generate either sets of ODEs or partial differential equations (PDEs). So basically \n`integrator` lets you solve any scientific or engineering problem which generates ODEs, which is a \nHUGE class of problems (FYI, finite element methods are used to solve sets of PDEs).\n\nFun fact: hundreds (or even thousands) of scientific problems had been formulated in the form of ODEs \nsince the time that Isaac Newton first invented calculus in the 1600's, but these problems remained \nintractable & unsolvable for over three centuries other than a very small handful that were amenable \nto a \"closed form solution\"; i.e., the ODEs could be solved analytically (i.e., via mathematical \nmanipulations). So there was this tragic dilemma; we could formulate these problems mathematically \nsince the 1600's - 1800's, but couldn't actually solve them. SAD! :disappointed:\n\nSo one of the primary drivers to create the first digital computers in the 1940's - 1960's was \nto solve ODEs. The space program, for example, would have been impossible without the numerical \nsolution of ODEs which represented the space flight trajectories, attitude, & control.  And before \nthe first digital computers, analog computers were used to solve ODEs back in the 1920's - 1940's.\n\nSo believe it or not, the first computers were developed and used to solve ODEs, not play League of \nLegends. :wink:\n\nThese algorithms are battle-tested and in some cases have been around for decades; Matlab and Octave \nare just relatively clean implementations of some of these algorithms, so I used them as the basis \nfor my Elixir versions.","ref":"readme.html#so-why-should-i-care-a-tool-to-solve-odes-wtf"},{"type":"extras","title":"Introduction to Integrator","doc":"# Introduction to Integrator\n\n```elixir\nMix.install([\n  {:integrator, \"~> 0.1.2\"},\n  {:kino_vega_lite, \"~> 0.1.7\"}\n])\n```","ref":"intro_to_integrator.html"},{"type":"extras","title":"Numerical Integration in Elixir - Introduction to Integrator","doc":"Numerical integration is easy with `Integrator`.  For example, let's integrate the \nVan der Pol equation `Integrator.SampleEqns.van_der_pol_fn` for 20 seconds:\n\n```elixir\nalias Integrator.SampleEqns\n\nt_initial = Nx.tensor(0.0, type: :f64)\nt_final = Nx.tensor(20.0, type: :f64)\nx_initial = Nx.tensor([2.0, 0.0], type: :f64)\nsolution = Integrator.integrate(&SampleEqns.van_der_pol_fn/2, [t_initial, t_final], x_initial)\n```\n\nNow you can plot the results via Kino:\n\n```elixir\nalias VegaLite, as: VL\n\ndata =\n  Enum.zip(solution.output_t, solution.output_x)\n  |> Enum.map(fn {t, x} ->\n    [\n      %{t: Nx.to_number(t), x: Nx.to_number(x[0]), x_value: \"x[0]\"},\n      %{t: Nx.to_number(t), x: Nx.to_number(x[1]), x_value: \"x[1]\"}\n    ]\n  end)\n  |> List.flatten()\n\nchart =\n  VL.new(width: 600, height: 400, title: \"Solution of van der Pol Equation (μ = 1) with ode45\")\n  |> VL.mark(:line, point: true, tooltip: true)\n  |> VL.encode_field(:x, \"t\", type: :quantitative)\n  |> VL.encode_field(:y, \"x\", type: :quantitative)\n  |> VL.encode_field(:color, \"x_value\", type: :nominal)\n  |> VL.data_from_values(data)\n  |> Kino.VegaLite.new()\n  |> Kino.render()\n```\n\nCompare this with the plot from the [Matlab ode45 manual page](https://www.mathworks.com/help/matlab/ref/ode45.html#bu3uj8b):\n\n![van der pol](https://www.mathworks.com/help/examples/matlab/win64/NonstiffVanDerPolEquationExample_01.png)","ref":"intro_to_integrator.html#numerical-integration-in-elixir"},{"type":"extras","title":"Interpolation and Fixed Times","doc":"# Interpolation and Fixed Times\n\n```elixir\nMix.install([\n  {:integrator, \"~> 0.1.2\"},\n  {:kino_vega_lite, \"~> 0.1.7\"}\n])\n```","ref":"interpolation_and_fixed_times.html"},{"type":"extras","title":"Interpolation - Interpolation and Fixed Times","doc":"By default, `refine: 4` for integrator `ode45`.  This means that four points are interpolated\nfor every solution of the ODE.\n\n```elixir\nalias Integrator.SampleEqns\n\nt_initial = Nx.tensor(0.0, type: :f64)\nt_final = Nx.tensor(20.0, type: :f64)\nx_initial = Nx.tensor([2.0, 0.0], type: :f64)\nsolution = Integrator.integrate(&SampleEqns.van_der_pol_fn/2, [t_initial, t_final], x_initial)\n```\n\nVisualizing `refine: 4`:\n\n```elixir\nalias VegaLite, as: VL\n\ndefmodule VanDerPol do\n  def plot(solution) do\n    data =\n      Enum.zip(solution.output_t, solution.output_x)\n      |> Enum.map(fn {t, x} ->\n        [\n          %{t: Nx.to_number(t), x: Nx.to_number(x[0]), x_value: \"x[0]\"},\n          %{t: Nx.to_number(t), x: Nx.to_number(x[1]), x_value: \"x[1]\"}\n        ]\n      end)\n      |> List.flatten()\n\n    VL.new(\n      width: 600,\n      height: 400,\n      title: \"Solution of van der Pol Equation (μ = 1) with ode45\"\n    )\n    |> VL.mark(:line, point: true, tooltip: true)\n    |> VL.encode_field(:x, \"t\", type: :quantitative)\n    |> VL.encode_field(:y, \"x\", type: :quantitative)\n    |> VL.encode_field(:color, \"x_value\", type: :nominal)\n    |> VL.data_from_values(data)\n    |> Kino.VegaLite.new()\n    |> Kino.render()\n  end\nend\n\nVanDerPol.plot(solution)\n```\n\nYou can turn off interpolation by setting `refine: 1`:\n\n```elixir\nt_initial = Nx.tensor(0.0, type: :f64)\nt_final = Nx.tensor(20.0, type: :f64)\nx_initial = Nx.tensor([2.0, 0.0], type: :f64)\nopts = [refine: 1]\n\nsolution =\n  Integrator.integrate(&SampleEqns.van_der_pol_fn/2, [t_initial, t_final], x_initial, opts)\n\nVanDerPol.plot(solution)\n```\n\nNote how much \"chunkier\" the plot is without the interpolated points.  These points are solely those from the Runge-Kutta simulation.  Note that these values can also be accessed by `solution.ode_t` and `solution.ode_x`.\n\n \n\nFinally, you can output data at fixed times. For example, let's print out data at `0.1` second intervals:\n\n```elixir\nt_initial = Nx.tensor(0.0, type: :f64)\nt_final = Nx.tensor(20.0, type: :f64)\nx_initial = Nx.tensor([2.0, 0.0], type: :f64)\nt_range = Nx.linspace(t_initial, t_final, n: 201, type: :f64)\nopts = [refine: 1]\nsolution = Integrator.integrate(&SampleEqns.van_der_pol_fn/2, t_range, x_initial, opts)\nVanDerPol.plot(solution)\n```","ref":"interpolation_and_fixed_times.html#interpolation"},{"type":"extras","title":"Output Functions","doc":"# Output Functions\n\n```elixir\nMix.install([\n  {:integrator, \"~> 0.1.2\"},\n  {:kino_vega_lite, \"~> 0.1.7\"}\n])\n```","ref":"output_functions.html"},{"type":"extras","title":"Usage - Output Functions","doc":"An output function lets you plot the results of an integration or simulation while it's occurring, or \nsend the data to an animation tool.  Let's see how that works.\n\nFirst, we'll need to set up an empty chart to receive the data:\n\n```elixir\nalias VegaLite, as: VL\n\nchart =\n  VL.new(\n    width: 600,\n    height: 400,\n    title: \"Solution of van der Pol Equation (μ = 1) with ode45\"\n  )\n  |> VL.mark(:line, point: true, tooltip: true)\n  |> VL.encode_field(:x, \"t\", type: :quantitative)\n  |> VL.encode_field(:y, \"x\", type: :quantitative)\n  |> VL.encode_field(:color, \"x_value\", type: :nominal)\n  |> Kino.VegaLite.new()\n  |> Kino.render()\n```\n\nNow, let's connect an output function to the simulation, and we'll inject a `Process.sleep(50)` on purpose so that the simulation takes a while. Watch the chart above while you run the simulation below.\n\n```elixir\nalias Integrator.SampleEqns\n\nt_initial = Nx.tensor(0.0, type: :f64)\nt_final = Nx.tensor(20.0, type: :f64)\nx_initial = Nx.tensor([2.0, 0.0], type: :f64)\n\noutput_fn = fn t, x ->\n  # t and x are lists of Nx tensors\n\n  Enum.zip(t, x)\n  |> Enum.map(fn {t, x} ->\n    [\n      %{t: Nx.to_number(t), x: Nx.to_number(x[0]), x_value: \"x[0]\"},\n      %{t: Nx.to_number(t), x: Nx.to_number(x[1]), x_value: \"x[1]\"}\n    ]\n  end)\n  |> List.flatten()\n  |> Enum.map(fn point ->\n    Kino.VegaLite.push(chart, point)\n\n    # Sleep on purpose to slow down the simulation:\n    Process.sleep(50)\n  end)\nend\n\nopts = [output_fn: output_fn, type: :f64]\n\nsolution =\n  Integrator.integrate(&SampleEqns.van_der_pol_fn/2, [t_initial, t_final], x_initial, opts)\n```","ref":"output_functions.html#usage"},{"type":"extras","title":"Event Functions","doc":"# Event Functions\n\n```elixir\nMix.install([\n  {:integrator, \"~> 0.1.2\"},\n  {:kino_vega_lite, \"~> 0.1.7\"}\n])\n```","ref":"event_functions.html"},{"type":"extras","title":"Usage - Event Functions","doc":"An event function lets you terminate a simulation based on some event (such as a collision).  For this example, we're going to mimic the [Matlab `ballode.m`](https://www.mathworks.com/help/matlab/math/ode-event-location.html) bouncing ball example. See also [here](https://github.com/kuo-courses/hw00/blob/master/ballode.m).\n\nThe equations of a bouncing ball are:\n\n$$ x_0 = x_1 $$\n\n$$ x_1 = - g  $$\n\nwhere $ g = 9.81 m/s^2 $.  Let's encode that in an Nx function:\n\n```elixir\nimport Nx.Defn\n\node_fn = fn _t, x ->\n  x0 = x[1]\n  x1 = -9.81\n  Nx.stack([x0, x1])\nend\n```\n\nThe follwing event function will detect when $ x_0 $ goes negative, and will return `:halt` in order to terminate the simulation:\n\n```elixir\nevent_fn = fn _t, x ->\n  value = Nx.to_number(x[0])\n  answer = if value   VL.mark(:line, point: true, tooltip: true)\n  |> VL.encode_field(:x, \"t\", type: :quantitative)\n  |> VL.encode_field(:y, \"x\", type: :quantitative)\n  |> VL.encode_field(:color, \"x_value\", type: :nominal)\n  |> Kino.VegaLite.new()\n\n# |> Kino.render()\n```\n\nThis output function will send the values of $ x_0 $ to the chart while the simulation is underway:\n\n```elixir\noutput_fn = fn t, x ->\n  Enum.zip(t, x)\n  |> Enum.map(fn {t, x} ->\n    [%{t: Nx.to_number(t), x: Nx.to_number(x[0]), x_value: \"x[0]\"}]\n  end)\n  |> List.flatten()\n  |> Enum.map(fn point ->\n    Kino.VegaLite.push(chart, point)\n  end)\nend\n```\n\nWe need to define a function which will determine what to do when transitions happen, which in our \ncase, are collisions between the ball and the ground. We'll reverse the direction of the ball, and\ndecrease its velocity by 10% (to account for bouncing).\n\n```elixir\ncoefficient_of_restitution = -0.9\n\ntransition_fn = fn t, x, _multi, opts ->\n  x1 = Nx.multiply(coefficient_of_restitution, x[1])\n  {:continue, t, Nx.stack([x[0], x1]), opts}\nend\n```\n\nThere's some recursive code in `Integrator.MultiIntegrator` that restarts the simulation when terminal \nevents are encountered.\n\n```elixir\nalias Integrator.MultiIntegrator\n\nt_initial = Nx.tensor(0.0, type: :f64)\nt_final = Nx.tensor(30.0, type: :f64)\nx_initial = Nx.tensor([0.0, 20.0], type: :f64)\nopts = [output_fn: output_fn]\n\nmulti_integrator =\n  MultiIntegrator.integrate(ode_fn, event_fn, transition_fn, t_initial, t_final, x_initial, opts)\n```\n\nCompare this plot with the version on the [Matlab page](https://www.mathworks.com/help/matlab/math/ode-event-location.html#bu7wjcg):\n\n![bouncing_ball](https://www.mathworks.com/help/examples/matlab/win64/SimpleEventLocationABouncingBallExample_01.png)","ref":"event_functions.html#usage"},{"type":"extras","title":"NonLinearEqnRoot Usage","doc":"# NonLinearEqnRoot Usage\n\n```elixir\n  Mix.install(\n  [\n    {:integrator, \"~> 0.1\"}\n  ]\n  )\n```","ref":"nonlinear_eqn_root.html"},{"type":"extras","title":"Finding a root - NonLinearEqnRoot Usage","doc":"(to be filled in later)","ref":"nonlinear_eqn_root.html#finding-a-root"}],"content_type":"text/markdown"}