searchNodes=[{"type":"module","title":"Integrator","doc":"A library for solving non-stiff ordinary differential equations (ODEs). Integrator uses either the Dormand-Prince 4/5 Runge Kutta algorithm, or the Bogacki-Shampine 2/3 Runge Kutta algorithm. It is intended that the user only needs to call Integrator.integrate/4 , and the Integrator.AdaptiveStepsize and Integrator.RungeKutta modules are only exposed for advanced users who want to use the underlying algorithms directly.","ref":"Integrator.html"},{"type":"function","title":"Integrator.integrate/4","doc":"Integrates an ODE function using either the Dormand-Prince45 method or the Bogacki-Shampine23 method. Options :initial_step - The initial stepsize. If not provided, a stepsize will be chosen automatically. Can be a float or a Nx tensor. :integrator - The integrator to use. Currently only :ode45 and :ode23 are supported, which correspond to Integrator.RungeKutta.DormandPrince45 and Integrator.RungeKutta.BogackiShampine23 , respectively. The default value is :ode45 . Additional Options Also see the options for these functions which are passed through: Integrator.NonLinearEqnRoot.find_zero/4 Integrator.AdaptiveStepsize.integrate/10","ref":"Integrator.html#integrate/4"},{"type":"function","title":"Integrator.options_schema_integrator_only/0","doc":"","ref":"Integrator.html#options_schema_integrator_only/0"},{"type":"module","title":"Integrator.AdaptiveStepsize","doc":"Integrates a set of ODEs with an adaptive timestep.","ref":"Integrator.AdaptiveStepsize.html"},{"type":"function","title":"Integrator.AdaptiveStepsize.abs_rel_norm_opts/1","doc":"Gets the default values used by the absolute-relative norm; e.g., abs_tol , rel_tol , and norm_control","ref":"Integrator.AdaptiveStepsize.html#abs_rel_norm_opts/1"},{"type":"function","title":"Integrator.AdaptiveStepsize.elapsed_time_ms/1","doc":"Returns the total elapsed time for the integration (in milleseconds)","ref":"Integrator.AdaptiveStepsize.html#elapsed_time_ms/1"},{"type":"function","title":"Integrator.AdaptiveStepsize.integrate/10","doc":"Integrates a set of ODEs. Options :abs_tol ( term/0 ) - The absolute tolerance used when computing the absolute relative norm. Defaults to 1.0e-06 in the Nx type that's been specified. :event_fn - A 2 arity function which determines whether an event has occured. If so, the integration is halted. The default value is nil . :max_number_of_errors ( integer/0 ) - The maximum number of permissible errors before the integration is halted. The default value is 5000 . :max_step ( term/0 ) - The default max time step. The default value is determined by the start and end times. :norm_control ( boolean/0 ) - Indicates whether norm control is to be used when computing the absolute relative norm. The default value is true . :output_fn - A 2 arity function which is called at each output point. The default value is nil . :refine - Indicates the number of additional interpolated points. 1 means no interpolation; 2 means one additional interpolated point; etc. :fixed_times means that the output times are fixed. The default value is 4 . :rel_tol ( term/0 ) - The relative tolerance used when computing the absolute relative norm. Defaults to 1.0e-03 in the Nx type that's been specified. :speed - :no_delay means to simulate as fast as possible. 1.0 means real time, 2.0 means twice as fast as real time, 0.5 means half as fast as real time, etc. The default value is :no_delay . :store_results? ( boolean/0 ) - Indicates whether or not to store the results of the integration. The default value is true . Additional Options Also see the options for the Integrator.NonLinearEqnRoot.find_zero/4 which are passed into integrate/10 . Originally adapted from the Octave integrate_adaptive.m See Wikipedia","ref":"Integrator.AdaptiveStepsize.html#integrate/10"},{"type":"function","title":"Integrator.AdaptiveStepsize.option_keys/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#option_keys/0"},{"type":"function","title":"Integrator.AdaptiveStepsize.options_schema/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#options_schema/0"},{"type":"function","title":"Integrator.AdaptiveStepsize.options_schema_adaptive_stepsize_only/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#options_schema_adaptive_stepsize_only/0"},{"type":"function","title":"Integrator.AdaptiveStepsize.starting_stepsize/7","doc":"Computes a good initial timestep for an ODE solver of order order using the algorithm described in the reference below. The input argument ode_fn , is the function describing the differential equations, t0 is the initial time, and x0 is the initial condition. abs_tol and rel_tol are the absolute and relative tolerance on the ODE integration. Originally based on the Octave starting_stepsize.m . Reference: E. Hairer, S.P. Norsett and G. Wanner, &quot;Solving Ordinary Differential Equations I: Nonstiff Problems&quot;, Springer.","ref":"Integrator.AdaptiveStepsize.html#starting_stepsize/7"},{"type":"type","title":"Integrator.AdaptiveStepsize.event_fn_t/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:event_fn_t/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.integration_status/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:integration_status/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.options_t/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:options_t/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.output_fn_t/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:output_fn_t/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.refine_strategy/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:refine_strategy/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.speed/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:speed/0"},{"type":"type","title":"Integrator.AdaptiveStepsize.t/0","doc":"","ref":"Integrator.AdaptiveStepsize.html#t:t/0"},{"type":"module","title":"Integrator.AdaptiveStepsize.ComputedStep","doc":"The results of the computation of an individual Runge-Kutta step","ref":"Integrator.AdaptiveStepsize.ComputedStep.html"},{"type":"type","title":"Integrator.AdaptiveStepsize.ComputedStep.t/0","doc":"","ref":"Integrator.AdaptiveStepsize.ComputedStep.html#t:t/0"},{"type":"module","title":"Integrator.MultiIntegrator","doc":"Integrates multiple simulations that are tied together somehow, such as a bouncing ball in the ballode.m example.","ref":"Integrator.MultiIntegrator.html"},{"type":"function","title":"Integrator.MultiIntegrator.all_output_data/2","doc":"Collates the simulation output from all of the integrations","ref":"Integrator.MultiIntegrator.html#all_output_data/2"},{"type":"function","title":"Integrator.MultiIntegrator.integrate/7","doc":"Integrates multiple times, with a transition function handling the junction between integrations Options See the options for these functions which are passed through: Integrator.NonLinearEqnRoot.find_zero/4 Integrator.AdaptiveStepsize.integrate/10 Integrator.integrate/4","ref":"Integrator.MultiIntegrator.html#integrate/7"},{"type":"type","title":"Integrator.MultiIntegrator.integration_status/0","doc":"","ref":"Integrator.MultiIntegrator.html#t:integration_status/0"},{"type":"type","title":"Integrator.MultiIntegrator.t/0","doc":"","ref":"Integrator.MultiIntegrator.html#t:t/0"},{"type":"type","title":"Integrator.MultiIntegrator.transition_fn_t/0","doc":"","ref":"Integrator.MultiIntegrator.html#t:transition_fn_t/0"},{"type":"module","title":"Integrator.NonLinearEqnRoot","doc":"Finds the roots (i.e., zeros) of a non-linear equation. Based on fzero.m from Octave. This is essentially the ACM algorithm 748: Enclosing Zeros of Continuous Functions due to Alefeld, Potra and Shi, ACM Transactions on Mathematical Software, Vol. 21, No. 3, September 1995. Although the workflow is the same, the structure of the algorithm has been transformed non-trivially; also, the algorithm has also been slightly modified.","ref":"Integrator.NonLinearEqnRoot.html"},{"type":"function","title":"Integrator.NonLinearEqnRoot.bracket_fx/1","doc":"","ref":"Integrator.NonLinearEqnRoot.html#bracket_fx/1"},{"type":"function","title":"Integrator.NonLinearEqnRoot.bracket_x/1","doc":"","ref":"Integrator.NonLinearEqnRoot.html#bracket_x/1"},{"type":"function","title":"Integrator.NonLinearEqnRoot.find_zero/4","doc":"Finds a zero for a function in an interval [a, b] (if the 2nd argument is a list) or in the vicinity of a (if the 2nd argument is a float). Options :max_iterations ( integer/0 ) - The maximum allowed number of iterations when finding a root. The default value is 1000 . :max_fn_eval_count ( integer/0 ) - The maximum allowed number of function evaluations when finding a root. The default value is 1000 . :type - The Nx type. The default value is :f64 . :machine_eps ( float/0 ) - The machine epsilon. Defaults to Nx.constants.epsilon/1 for this Nx type. :tolerance ( float/0 ) - The tolerance for the convergence when finding a root. Defaults to Nx.Constants.epsilon/1 for this Nx type. :nonlinear_eqn_root_output_fn ( term/0 ) - An output function to call so intermediate results can be retrieved when finding a root. The default value is nil .","ref":"Integrator.NonLinearEqnRoot.html#find_zero/4"},{"type":"function","title":"Integrator.NonLinearEqnRoot.option_keys/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#option_keys/0"},{"type":"function","title":"Integrator.NonLinearEqnRoot.options_schema/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#options_schema/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.convergence_status/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:convergence_status/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.interpolation_type/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:interpolation_type/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.iter_type/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:iter_type/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.options_t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:options_t/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.output_fn_t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:output_fn_t/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.search_for_2nd_point_t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:search_for_2nd_point_t/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:t/0"},{"type":"type","title":"Integrator.NonLinearEqnRoot.zero_fn_t/0","doc":"","ref":"Integrator.NonLinearEqnRoot.html#t:zero_fn_t/0"},{"type":"behaviour","title":"Integrator.RungeKutta","doc":"A behaviour that Runge-Kutta algorithms must implement. Currently, Integrator.DormandPrince45 and Integrator.BogackiShampine23 implement this behaviour. See the list of Runge-Kutta methods","ref":"Integrator.RungeKutta.html"},{"type":"callback","title":"Integrator.RungeKutta.integrate/6","doc":"Integrates an ODE function","ref":"Integrator.RungeKutta.html#c:integrate/6"},{"type":"callback","title":"Integrator.RungeKutta.interpolate/4","doc":"Interpolates using the method that is suitable for this particular Runge-Kutta method","ref":"Integrator.RungeKutta.html#c:interpolate/4"},{"type":"callback","title":"Integrator.RungeKutta.order/0","doc":"The order of this Runge-Kutta method","ref":"Integrator.RungeKutta.html#c:order/0"},{"type":"type","title":"Integrator.RungeKutta.interpolate_fn_t/0","doc":"","ref":"Integrator.RungeKutta.html#t:interpolate_fn_t/0"},{"type":"type","title":"Integrator.RungeKutta.ode_fn_t/0","doc":"","ref":"Integrator.RungeKutta.html#t:ode_fn_t/0"},{"type":"type","title":"Integrator.RungeKutta.stepper_fn_t/0","doc":"","ref":"Integrator.RungeKutta.html#t:stepper_fn_t/0"},{"type":"module","title":"Integrator.RungeKutta.BogackiShampine23","doc":"Bogacki-Shampine method of third order. For the definition of this method see Wikipedia Originally based on Octave runge_kutta_23.m","ref":"Integrator.RungeKutta.BogackiShampine23.html"},{"type":"function","title":"Integrator.RungeKutta.BogackiShampine23.integrate/6","doc":"Solves a set of non-stiff Ordinary Differential Equations (non-stiff ODEs) with the well-known explicit Bogacki-Shampine method of order 3. See Wikipedia here and here","ref":"Integrator.RungeKutta.BogackiShampine23.html#integrate/6"},{"type":"function","title":"Integrator.RungeKutta.BogackiShampine23.interpolate/4","doc":"Performs a Hermite cubic interpolation when using BogackiShampine23 via Utils.hermite_cubic_interpolation/4","ref":"Integrator.RungeKutta.BogackiShampine23.html#interpolate/4"},{"type":"function","title":"Integrator.RungeKutta.BogackiShampine23.order/0","doc":"Returns the order of this Runge-Kutta method (which is 3)","ref":"Integrator.RungeKutta.BogackiShampine23.html#order/0"},{"type":"module","title":"Integrator.RungeKutta.DormandPrince45","doc":"Integrates and interpolates a system of ODEs with a given initial condition x from t to t + dt with the Dormand-Prince method. For the definition of this method see Wikipedia . Originally based on runge_kutta_45_dorpri.m from Octave. It uses six function evaluations to calculate fourth and fifth-order accurate solutions. The difference between these solutions is then taken to be the error of the (fourth-order) solution. This error estimate is very convenient for adaptive stepsize integration algorithms. The Dormand–Prince method has seven stages, but it uses only six function evaluations per step because it has the FSAL (First Same As Last) property: the last stage is evaluated at the same point as the first stage of the next step. Dormand and Prince chose the coefficients of their method to minimize the error of the fifth-order solution. This is the main difference with the Fehlberg method , which was constructed so that the fourth-order solution has a small error. See the Octave docs for ode45","ref":"Integrator.RungeKutta.DormandPrince45.html"},{"type":"function","title":"Integrator.RungeKutta.DormandPrince45.integrate/6","doc":"Integrates a system of ODEs with [Dormand-Prince]]( http://en.wikipedia.org/wiki/Dormand%E2%80%93Prince_method ). Reference: Hairer, Ernst; Nørsett, Syvert Paul; Wanner, Gerhard (2008), Solving ordinary differential equations I: Nonstiff problems, Berlin, New York: Springer-Verlag, ISBN 978-3-540-56670-0","ref":"Integrator.RungeKutta.DormandPrince45.html#integrate/6"},{"type":"function","title":"Integrator.RungeKutta.DormandPrince45.interpolate/4","doc":"Performs a 4th order Hermite interpolation when interpolating with DormandPrince45 using Utils.hermite_quartic_interpolation/4","ref":"Integrator.RungeKutta.DormandPrince45.html#interpolate/4"},{"type":"function","title":"Integrator.RungeKutta.DormandPrince45.order/0","doc":"Returns the order of this Runge-Kutta method (which is 5)","ref":"Integrator.RungeKutta.DormandPrince45.html#order/0"},{"type":"module","title":"Integrator.SampleEqns","doc":"Functions to be used in testing","ref":"Integrator.SampleEqns.html"},{"type":"function","title":"Integrator.SampleEqns.euler_equations/2","doc":"The Euler equations of a rigid body without external forces. This is a standard test problem proposed by Krogh for solvers intended for nonstiff problems [see below]. Based on &quot;rigidode.m&quot; from Matlab/Octave. The analytical solutions are Jacobian elliptic functions. See rigidode.m in Matlab. Shampine, L. F., and M. K. Gordon, Computer Solution of Ordinary Differential Equations, W.H. Freeman &amp; Co., 1975","ref":"Integrator.SampleEqns.html#euler_equations/2"},{"type":"function","title":"Integrator.SampleEqns.falling_particle/2","doc":"Simulates a point mass or particle falling through pass affected by gravity. Used for comparisons with the Matlab/Octave ballode.m routine","ref":"Integrator.SampleEqns.html#falling_particle/2"},{"type":"function","title":"Integrator.SampleEqns.van_der_pol_fn/2","doc":"From octave b fvdp = @(t,x) [x(2); (1 - x(1)^2) * x(2) - x(1)];","ref":"Integrator.SampleEqns.html#van_der_pol_fn/2"},{"type":"module","title":"Integrator.Utils","doc":"Various utility functions used in Integrator","ref":"Integrator.Utils.html"},{"type":"function","title":"Integrator.Utils.columns_as_list/3","doc":"Returns the columns of a tensor as a list of vector tensors E.g., a tensor of the form: ~M[ 1 2 3 4 5 6 7 8 9 10 11 12 ]s8 will return [ ~V[ 1 5 9 ]s8 , ~V[ 2 6 10 ]s8 , ~V[ 3 7 11 ]s8 , ~V[ 4 8 12 ]s8 ]","ref":"Integrator.Utils.html#columns_as_list/3"},{"type":"function","title":"Integrator.Utils.epsilon/1","doc":"Gets the epsilon for a particular Nx type","ref":"Integrator.Utils.html#epsilon/1"},{"type":"function","title":"Integrator.Utils.epsilon_nx/1","doc":"","ref":"Integrator.Utils.html#epsilon_nx/1"},{"type":"function","title":"Integrator.Utils.hermite_cubic_interpolation/4","doc":"Performs a 3rd order Hermite interpolation. Adapted from function hermite_cubic_interpolation in runge_kutta_interpolate.m See Wikipedia","ref":"Integrator.Utils.html#hermite_cubic_interpolation/4"},{"type":"function","title":"Integrator.Utils.hermite_quartic_interpolation/4","doc":"Performs a 4th order Hermite interpolation. Used by an ODE solver to interpolate the solution at the time t_out . As proposed by Shampine in Lawrence, Shampine, &quot;Some Practical Runge-Kutta Formulas&quot;, 1986. See hermite_quartic_interpolation function in Octave .","ref":"Integrator.Utils.html#hermite_quartic_interpolation/4"},{"type":"function","title":"Integrator.Utils.kahan_sum/3","doc":"Implements the Kahan summation algorithm, also known as compensated summation. Based on this code in Octave . This is really a private function, but is made public so the docs are visible. The algorithm significantly reduces the numerical error in the total obtained by adding a sequence of finite precision floating point numbers compared to the straightforward approach. For more details see this Wikipedia entry . This function is called by AdaptiveStepsize.integrate to better catch equality comparisons. The first input argument is the variable that will contain the summation. This variable is also returned as the first output argument in order to reuse it in subsequent calls to kahan_sum/3 function. The second input argument contains the compensation term and is returned as the second output argument so that it can be reused in future calls of the same summation. The third input argument term is the variable to be added to sum .","ref":"Integrator.Utils.html#kahan_sum/3"},{"type":"function","title":"Integrator.Utils.sign/1","doc":"Returns the sign of the tensor as -1 or 1 (or 0 for zero tensors)","ref":"Integrator.Utils.html#sign/1"},{"type":"function","title":"Integrator.Utils.unique/1","doc":"Given a list of elements, create a new list with only the unique elements from the list","ref":"Integrator.Utils.html#unique/1"},{"type":"function","title":"Integrator.Utils.vector_as_list/1","doc":"Converts a Nx vector into a list of 1-D tensors Is there an existing Nx way to do this? If so, swap the usage of this function and then delete this Note that vector |&gt; Nx . as_list ( ) |&gt; Enum . map ( &amp; Nx . tensor ( &amp;1 , type : Nx . type ( vector ) ) ) seems to introduce potential precision issues","ref":"Integrator.Utils.html#vector_as_list/1"},{"type":"extras","title":"Integrator","doc":"A numerical integrator written in Elixir for the solution of sets of non-stiff ordinary differential equations (ODEs).","ref":"readme.html"},{"type":"extras","title":"Integrator - Installation","doc":"The package can be installed by adding integrator to your list of dependencies in mix.exs : def deps do [ { :integrator , &quot;~&gt; 0.1&quot; } , ] end The docs can be found at https://hexdocs.pm/integrator .","ref":"readme.html#installation"},{"type":"extras","title":"Integrator - Description","doc":"Two integrator options are available; ode45 which is an adaptation of the Octave ode45 and Matlab ode45 . The ode45 integrator utilizes the Dormand-Prince 4th/5th order Runge Kutta algorithm. ode23 is an adaptation of the Octave ode23 and Matlab ode23 The ode23 integrator uses the Bogacki-Shampine 3rd order Runge Kutta algorithm. Both ode45 (which is the default integrator option) and ode23 utilize an adaptive stepsize algorithm for computing the integration time step. The time step is computed based on the satisfaction of a required error tolerance. This library heavily leverages Elixir Nx ; many thanks to the creators of Nx , as without it this library would not have been possible. The GNU Octave code was also used heavily for inspiration and was used to generate numerical test cases for the Elixir versions of the algorithms. Many thanks to John W. Eaton for his tremendous work on Octave. Integrator has been tested extensively during its development, and has a large and growing test suite.","ref":"readme.html#description"},{"type":"extras","title":"Integrator - Usage","doc":"See the Livebook guides for detailed examples of usage. As a simple example, you can integrate the Van der Pol equation as defined in Integrator.SampleEqns.van_der_pol_fn/2 from time 0 to 20 with an intial x value of [0, 1] via: t_initial = 0.0 t_final = 20.0 x_initial = Nx . tensor ( [ 0.0 , 1.0 ] ) solution = Integrator . integrate ( &amp; SampleEqns . van_der_pol_fn / 2 , [ t_initial , t_final ] , x_initial ) Then, solution.output_t contains a list of output times, and solution.output_x contains a list of values of x at these corresponding times. Options exist for: outputting simulation results dynamically via an output function (for applications such as plotting dynamically, or for animating while the simulation is underway) generating simulation output at fixed times (such as at t = 0.1, 0.2, 0.3 , etc.) interpolating intermediate points via quartic Hermite interpolation (for ode45 ) or via cubic Hermite interpolation (for ode23 ) detecting termination events (such as collisions); see the Livebooks for details. increasing the simulation fidelity (at the expense of simulation time) via absolute tolerance and relative tolerance settings","ref":"readme.html#usage"},{"type":"extras","title":"Introduction to Integrator","doc":"Mix . install ( [ { :integrator , &quot;~&gt; 0.1&quot; } , { :kino_vega_lite , &quot;~&gt; 0.1.7&quot; } ] )","ref":"intro_to_integrator.html"},{"type":"extras","title":"Introduction to Integrator - Numerical Integration in Elixir","doc":"Numerical integration is easy with Integrator . For example, let's integrate the Van der Pol equation Integrator.SampleEqns.van_der_pol_fn for 20 seconds: alias Integrator.SampleEqns t_initial = 0.0 t_final = 20.0 x_initial = Nx . tensor ( [ 2.0 , 0.0 ] ) solution = Integrator . integrate ( &amp; SampleEqns . van_der_pol_fn / 2 , [ t_initial , t_final ] , x_initial ) Now you can plot the results via Kino: alias VegaLite , as : VL data = Enum . zip ( solution . output_t , solution . output_x ) |&gt; Enum . map ( fn { t , x } -&gt; [ %{ t : Nx . to_number ( t ) , x : Nx . to_number ( x [ 0 ] ) , x_value : &quot;x[0]&quot; } , %{ t : Nx . to_number ( t ) , x : Nx . to_number ( x [ 1 ] ) , x_value : &quot;x[1]&quot; } ] end ) |&gt; List . flatten ( ) chart = VL . new ( width : 600 , height : 400 , title : &quot;Solution of van der Pol Equation (μ = 1) with ode45&quot; ) |&gt; VL . mark ( :line , point : true , tooltip : true ) |&gt; VL . encode_field ( :x , &quot;t&quot; , type : :quantitative ) |&gt; VL . encode_field ( :y , &quot;x&quot; , type : :quantitative ) |&gt; VL . encode_field ( :color , &quot;x_value&quot; , type : :nominal ) |&gt; VL . data_from_values ( data ) |&gt; Kino.VegaLite . new ( ) |&gt; Kino . render ( ) Compare this with the plot from the Matlab ode45 manual page :","ref":"intro_to_integrator.html#numerical-integration-in-elixir"},{"type":"extras","title":"Interpolation and Fixed Times","doc":"Mix . install ( [ { :integrator , &quot;~&gt; 0.1&quot; } , { :kino_vega_lite , &quot;~&gt; 0.1.7&quot; } ] )","ref":"interpolation_and_fixed_times.html"},{"type":"extras","title":"Interpolation and Fixed Times - Interpolation","doc":"By default, refine: 4 for integrator ode45 . This means that four points are interpolated for every solution of the ODE. alias Integrator.SampleEqns t_initial = 0.0 t_final = 20.0 x_initial = Nx . tensor ( [ 2.0 , 0.0 ] ) solution = Integrator . integrate ( &amp; SampleEqns . van_der_pol_fn / 2 , [ t_initial , t_final ] , x_initial ) Visualizing refine: 4 : alias VegaLite , as : VL defmodule VanDerPol do def plot ( solution ) do data = Enum . zip ( solution . output_t , solution . output_x ) |&gt; Enum . map ( fn { t , x } -&gt; [ %{ t : Nx . to_number ( t ) , x : Nx . to_number ( x [ 0 ] ) , x_value : &quot;x[0]&quot; } , %{ t : Nx . to_number ( t ) , x : Nx . to_number ( x [ 1 ] ) , x_value : &quot;x[1]&quot; } ] end ) |&gt; List . flatten ( ) VL . new ( width : 600 , height : 400 , title : &quot;Solution of van der Pol Equation (μ = 1) with ode45&quot; ) |&gt; VL . mark ( :line , point : true , tooltip : true ) |&gt; VL . encode_field ( :x , &quot;t&quot; , type : :quantitative ) |&gt; VL . encode_field ( :y , &quot;x&quot; , type : :quantitative ) |&gt; VL . encode_field ( :color , &quot;x_value&quot; , type : :nominal ) |&gt; VL . data_from_values ( data ) |&gt; Kino.VegaLite . new ( ) |&gt; Kino . render ( ) end end VanDerPol . plot ( solution ) You can turn off interpolation by setting refine: 1 : t_initial = 0.0 t_final = 20.0 x_initial = Nx . tensor ( [ 2.0 , 0.0 ] ) opts = [ refine : 1 ] solution = Integrator . integrate ( &amp; SampleEqns . van_der_pol_fn / 2 , [ t_initial , t_final ] , x_initial , opts ) VanDerPol . plot ( solution ) Note how much &quot;chunkier&quot; the plot is without the interpolated points. These points are solely those from the Runge-Kutta simulation. Note that these values can also be accessed by solution.ode_t and solution.ode_x . Finally, you can output data at fixed times. For example, let's print out data at 0.1 second intervals: t_initial = 0.0 t_final = 20.0 x_initial = Nx . tensor ( [ 2.0 , 0.0 ] ) t_range = Nx . linspace ( t_initial , t_final , n : 201 ) opts = [ refine : 1 ] solution = Integrator . integrate ( &amp; SampleEqns . van_der_pol_fn / 2 , t_range , x_initial , opts ) VanDerPol . plot ( solution )","ref":"interpolation_and_fixed_times.html#interpolation"},{"type":"extras","title":"Output Functions","doc":"Mix . install ( [ { :integrator , &quot;~&gt; 0.1&quot; } , { :kino_vega_lite , &quot;~&gt; 0.1.7&quot; } ] )","ref":"output_functions.html"},{"type":"extras","title":"Output Functions - Usage","doc":"An output function lets you plot the results of an integration or simulation while it's occurring, or send the data to an animation tool. Let's see how that works. First, we'll need to set up an empty chart to receive the data: alias VegaLite , as : VL chart = VL . new ( width : 600 , height : 400 , title : &quot;Solution of van der Pol Equation (μ = 1) with ode45&quot; ) |&gt; VL . mark ( :line , point : true , tooltip : true ) |&gt; VL . encode_field ( :x , &quot;t&quot; , type : :quantitative ) |&gt; VL . encode_field ( :y , &quot;x&quot; , type : :quantitative ) |&gt; VL . encode_field ( :color , &quot;x_value&quot; , type : :nominal ) |&gt; Kino.VegaLite . new ( ) |&gt; Kino . render ( ) Now, let's connect an output function to the simulation, and we'll inject a Process.sleep(50) on purpose so that the simulation takes a while. Watch the chart above while you run the simulation below. alias Integrator.SampleEqns t_initial = 0.0 t_final = 20.0 x_initial = Nx . tensor ( [ 2.0 , 0.0 ] ) output_fn = fn t , x -&gt; # t and x are lists of Nx tensors Enum . zip ( t , x ) |&gt; Enum . map ( fn { t , x } -&gt; [ %{ t : Nx . to_number ( t ) , x : Nx . to_number ( x [ 0 ] ) , x_value : &quot;x[0]&quot; } , %{ t : Nx . to_number ( t ) , x : Nx . to_number ( x [ 1 ] ) , x_value : &quot;x[1]&quot; } ] end ) |&gt; List . flatten ( ) |&gt; Enum . map ( fn point -&gt; Kino.VegaLite . push ( chart , point ) # Sleep on purpose to slow down the simulation: Process . sleep ( 50 ) end ) end opts = [ output_fn : output_fn ] solution = Integrator . integrate ( &amp; SampleEqns . van_der_pol_fn / 2 , [ t_initial , t_final ] , x_initial , opts )","ref":"output_functions.html#usage"},{"type":"extras","title":"Event Functions","doc":"Mix . install ( [ { :integrator , &quot;~&gt; 0.1&quot; } , { :kino_vega_lite , &quot;~&gt; 0.1.7&quot; } ] )","ref":"event_functions.html"},{"type":"extras","title":"Event Functions - Usage","doc":"An event function lets you terminate a simulation based on some event (such as a collision). For this example, we're going to mimic the Matlab ballode.m bouncing ball example. See also here . The equations of a bouncing ball are: $$ x_0 = x_1 $$ $$ x_1 = - g $$ where $ g = 9.81 m/s^2 $. Let's encode that in an Nx function: import Nx.Defn ode_fn = fn _t , x -&gt; x0 = x [ 1 ] x1 = - 9.81 Nx . stack ( [ x0 , x1 ] ) end The follwing event function will detect when $ x_0 $ goes negative, and will return :halt in order to terminate the simulation: event_fn = fn _t , x -&gt; value = Nx . to_number ( x [ 0 ] ) answer = if value &lt;= 0.0 , do : :halt , else : :continue { answer , value } end Create an empty chart to receive the data: alias VegaLite , as : VL chart = VL . new ( width : 600 , height : 400 , title : &quot;Bouncing Ball&quot; ) |&gt; VL . mark ( :line , point : true , tooltip : true ) |&gt; VL . encode_field ( :x , &quot;t&quot; , type : :quantitative ) |&gt; VL . encode_field ( :y , &quot;x&quot; , type : :quantitative ) |&gt; VL . encode_field ( :color , &quot;x_value&quot; , type : :nominal ) |&gt; Kino.VegaLite . new ( ) # |&gt; Kino.render() This output function will send the values of $ x_0 $ to the chart while the simulation is underway: output_fn = fn t , x -&gt; Enum . zip ( t , x ) |&gt; Enum . map ( fn { t , x } -&gt; [ %{ t : Nx . to_number ( t ) , x : Nx . to_number ( x [ 0 ] ) , x_value : &quot;x[0]&quot; } ] end ) |&gt; List . flatten ( ) |&gt; Enum . map ( fn point -&gt; Kino.VegaLite . push ( chart , point ) end ) end We need to define a function which will determine what to do when transitions happen, which in our case, are collisions between the ball and the ground. We'll reverse the direction of the ball, and decrease its velocity by 10% (to account for bouncing). coefficient_of_restitution = - 0.9 transition_fn = fn t , x , _multi , opts -&gt; x1 = Nx . multiply ( coefficient_of_restitution , x [ 1 ] ) { :continue , t , Nx . stack ( [ x [ 0 ] , x1 ] ) , opts } end There's some recursive code in Integrator.MultiIntegrator that restarts the simulation when terminal events are encountered. alias Integrator.MultiIntegrator t_initial = Nx . tensor ( 0.0 , type : :f64 ) t_final = Nx . tensor ( 30.0 , type : :f64 ) x_initial = Nx . tensor ( [ 0.0 , 20.0 ] , type : :f64 ) opts = [ output_fn : output_fn ] multi_integrator = MultiIntegrator . integrate ( ode_fn , event_fn , transition_fn , t_initial , t_final , x_initial , opts ) Compare this plot with the version on the Matlab page :","ref":"event_functions.html#usage"},{"type":"extras","title":"NonLinearEqnRoot Usage","doc":"Mix . install ( [ { :integrator , &quot;~&gt; 0.1&quot; } ] )","ref":"nonlinear_eqn_root.html"},{"type":"extras","title":"NonLinearEqnRoot Usage - Finding a root","doc":"","ref":"nonlinear_eqn_root.html#finding-a-root"}]